﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Problems.VehicleRouting;

namespace HeuristicLab.PDPSimulation.Operators {
  [Item("DynPotvinPDExchangeMoveMaker", "Peforms the exchange move on a given PDP encoding and updates the quality.")]
  [StorableClass]
  public class DynPotvinPDExchangeMoveMaker : PotvinMoveMaker, IDynPotvinPDExchangeMoveOperator, IMoveMaker {
    public ILookupParameter<DynPotvinPDExchangeMove> PDExchangeMoveParameter {
      get { return (ILookupParameter<DynPotvinPDExchangeMove>)Parameters["DynPotvinPDExchangeMove"]; }
    }

    public override ILookupParameter VRPMoveParameter {
      get { return PDExchangeMoveParameter; }
    }

    [StorableConstructor]
    private DynPotvinPDExchangeMoveMaker(bool deserializing) : base(deserializing) { }

    public DynPotvinPDExchangeMoveMaker()
      : base() {
        Parameters.Add(new LookupParameter<DynPotvinPDExchangeMove>("DynPotvinPDExchangeMove", "The moves that should be made."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DynPotvinPDExchangeMoveMaker(this, cloner);
    }

    protected DynPotvinPDExchangeMoveMaker(DynPotvinPDExchangeMoveMaker original, Cloner cloner)
      : base(original, cloner) {
    }

    public static void Apply(PotvinEncoding solution, DynPotvinPDExchangeMove move, IVRPProblemInstance problemInstance) {
      if (move.Tour >= solution.Tours.Count)
        solution.Tours.Add(new Tour());
      Tour tour = solution.Tours[move.Tour];

      Tour oldTour = solution.Tours.Find(t => t.Stops.Contains(move.City));
      oldTour.Stops.Remove(move.City);

      if (problemInstance is IPickupAndDeliveryProblemInstance) {
        IPickupAndDeliveryProblemInstance pdp = problemInstance as IPickupAndDeliveryProblemInstance;

        Tour oldTour2;
        int location = pdp.GetPickupDeliveryLocation(move.City);
        if (location > 0 && location != move.City) {
          oldTour2 = solution.Tours.Find(t => t.Stops.Contains(location));
          oldTour2.Stops.Remove(location);
        }

        location = pdp.GetPickupDeliveryLocation(move.Replaced);
        if (location > 0 && location != move.Replaced) {
          oldTour2 = solution.Tours.Find(t => t.Stops.Contains(location));
          oldTour2.Stops.Remove(location);
        }

        tour.Stops.Remove(move.Replaced);

        if (move.City != pdp.GetPickupDeliveryLocation(move.City))
          DynPotvinPDRearrangeMoveMaker.InsertPair(solution, tour, move.City, pdp.GetPickupDeliveryLocation(move.City), problemInstance);
        else {
          int place = solution.FindBestInsertionPlace(tour, move.City);
          tour.Stops.Insert(place, move.City);
        }

        if (move.Replaced != pdp.GetPickupDeliveryLocation(move.Replaced))
          DynPotvinPDRearrangeMoveMaker.InsertPair(solution, oldTour, move.Replaced, pdp.GetPickupDeliveryLocation(move.Replaced), problemInstance);
        else {
          int place = solution.FindBestInsertionPlace(oldTour, move.Replaced);
          oldTour.Stops.Insert(place, move.Replaced);
        }
      } else {
        tour.Stops.Remove(move.Replaced);

        int place = solution.FindBestInsertionPlace(tour, move.City);
        tour.Stops.Insert(place, move.City);

        place = solution.FindBestInsertionPlace(oldTour, move.Replaced);
        oldTour.Stops.Insert(place, move.Replaced);
      }
    }

    protected override void PerformMove() {
      DynPotvinPDExchangeMove move = PDExchangeMoveParameter.ActualValue;

      PotvinEncoding newSolution = move.Individual.Clone() as PotvinEncoding;
      Apply(newSolution, move, ProblemInstance);
      newSolution.Repair();
      VRPToursParameter.ActualValue = newSolution;
    }
  }
}
