﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Parameters;
using System.Collections.Generic;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Problems.VehicleRouting.Encodings.General;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Encodings.PermutationEncoding;

namespace HeuristicLab.PDPSimulation.Operators {
  [Item("DynPotvinPDShiftMultiMoveGenerator", "Generates shift moves from a given PDP encoding.")]
  [StorableClass]
  public sealed class DynPotvinPDShiftMultiMoveGenerator : DynPotvinPDShiftMoveGenerator, IMultiMoveGenerator, IMultiVRPMoveGenerator {
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public IValueLookupParameter<IntValue> SampleSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["SampleSize"]; }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DynPotvinPDShiftMultiMoveGenerator(this, cloner);
    }

    [StorableConstructor]
    private DynPotvinPDShiftMultiMoveGenerator(bool deserializing) : base(deserializing) { }

    public DynPotvinPDShiftMultiMoveGenerator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator."));
      Parameters.Add(new ValueLookupParameter<IntValue>("SampleSize", "The number of moves to generate."));
    }

    private DynPotvinPDShiftMultiMoveGenerator(DynPotvinPDShiftMultiMoveGenerator original, Cloner cloner)
      : base(original, cloner) {
    }

    protected override DynPotvinPDShiftMove[] GenerateMoves(PotvinEncoding individual, IVRPProblemInstance problemInstance) {
      List<DynPotvinPDShiftMove> result = new List<DynPotvinPDShiftMove>();
      IRandom rand = RandomParameter.ActualValue;
      double samples = Math.Min(100, SampleSizeParameter.ActualValue.Value);

      List<int> cities = new List<int>();
      IPickupAndDeliveryProblemInstance pdp = problemInstance as IPickupAndDeliveryProblemInstance;
      for (int i = 1; i <= individual.Cities; i++) {
        if (pdp == null || pdp.GetDemand(i) >= 0 || pdp.GetPickupDeliveryLocation(i) <= 0 || pdp.GetPickupDeliveryLocation(i) == i)
          cities.Add(i);
      }

      if (cities.Count >= 1) {
        int max = individual.Tours.Count;
        if (individual.Tours.Count >= problemInstance.Vehicles.Value)
          max = max - 1;

        if (max > 0) {
          Dictionary<int, int> tours = new Dictionary<int, int>();
          for (int tourIdx = 0; tourIdx < individual.Tours.Count; tourIdx++) {
            for (int stop = 0; stop < individual.Tours[tourIdx].Stops.Count; stop++) {
              int city = individual.Tours[tourIdx].Stops[stop];
              tours[city] = tourIdx;
            }
          }

          Permutation perm = new Permutation(PermutationTypes.Absolute, cities.Count * max);
          perm.Randomize(rand);

          samples = Math.Max(1, (int)(perm.Length * (samples / 100.0)));

          for (int i = 0; i < Math.Min(perm.Length, samples); i++) {
            int city = cities[perm[i] / max];
            int oldTourIndex = tours[city];

            int newTourIndex = perm[i] % max;
            if (newTourIndex >= oldTourIndex)
              newTourIndex++;

            var move = new DynPotvinPDShiftMove(city, oldTourIndex, newTourIndex, individual);
            result.Add(move);
          }
        }
      }

      return result.ToArray();
    }
  }
}
